# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from .problems import Clustering as Clustering
from .problems import Perceptron as Perceptron
from .problems import SammonMapping as SammonMapping
from .problems import Landscape as Landscape
